/***************************************************
 *                                                 *
 *    Macro Assembler			           *
 *                                                 *
 *    Copyright 1989    Richard P. Cornwell        *
 *    Developmental Code not for redistubution     *
 *                                                 *
 *    Written by Richard Cornwell                  *
 *               3/20/86                           *
 *                                                 *
 ***************************************************/

#include "macro.h"

#define REGFIELD(x)	((x>>8)&0377)

#define LOCK	000001		/* Lock prefix */
#define REP	000002		/* Repeat prefix */
#define SEG	000004		/* Segment override */
#define WD	000010		/* Word instruction */
#define LG	000020		/* Long instruction */
#define ADDR16	000040		/* 16 bit addressing mode used */
#define ADDR32	000100		/* 32 bit addressing mode used */
#define TWOB	000200		/* Two byte opcode */
#define IMMED	000400		/* Immedate instruction */
#define MODRM	001000		/* Uses mod/rm byte i.e. reg only operand */
#define SIMMED	002000		/* One byte sign extended immedate operand */
#define FAR	004000		/* Far address, get segment from immedate */
#define SJMP	010000		/* Generate a short jump over the next op */

#define SEGCS	1		/* CS register */
#define SEGSS	2		/* SS register */
#define SEGDS	3		/* DS register */
#define SEGNONE	010		/* No segment overide */

#define REGAX	0
#define REGBX	3
#define REGCX	1
#define REGDX	2
#define REGSI	6
#define REGDI	7
#define REGBP	5
#define REGSP	4

void genea();

/* es,   cs,   ss,   ds,   fs,   gs */
char            segover[8] = {0x26, 0x2e, 0x36, 0x3e, 0x64, 0x65, 0, 0};
char            segpush[8] = {0x06, 0x0e, 0x16, 0x1e, 0xa0, 0xa8, 0, 0};
int             mmode = 1;
/* oprsize [(right<<2)|left] */
		/*	  4210 */
long oprsize[] = {	0x40210,		/* right = 0 left = 0 */
			0x00011,		/* right = 0 left = 1 */
			0x00202,		/* right = 0 left = 2 */
			0x00000,		/* right = 0 left = 3 */
			0x40004,		/* right = 0 left = 4 */
			0, 0, 0, 	/* left = 5, 6, 7 */
			0x00011,		/* right = 1 left = 0 */
			0x00011,		/* right = 1 left = 1 */
			0x00000,		/* right = 1 left = 2 */
			0x00000,		/* right = 1 left = 3 */
			0x00000,		/* right = 1 left = 4 */
			0, 0, 0, 	/* left = 5, 6, 7 */
			0x00202,		/* right = 2 left = 0 */
			0x00000,		/* right = 2 left = 1 */
			0x00202,		/* right = 2 left = 2 */
			0x00000,		/* right = 2 left = 3 */
			0x00000,		/* right = 2 left = 4 */
			0, 0, 0, 	/* left = 5, 6, 7 */
			0x00000,		/* right = 3 left = 0 */
			0x00000,		/* right = 3 left = 1 */
			0x00000,		/* right = 3 left = 2 */
			0x00000,		/* right = 3 left = 3 */
			0x00000,		/* right = 3 left = 4 */
			0, 0, 0, 	/* left = 5, 6, 7 */
			0x40004,		/* right = 4 left = 0 */
			0x00000,		/* right = 4 left = 1 */
			0x00000,		/* right = 4 left = 2 */
			0x00000,		/* right = 4 left = 3 */
			0x40004		/* right = 4 left = 4 */
};

void
machine(ptr)
	struct opcodes *ptr;
{
	register int    i, sz, lsz, rsz;
	int             f;
	int             base;
	static short int flag = 0;	/* Prefix indicators */
	static char     repeat;	/* Value of repeat opcode */
	static char     segment;/* Segment */
	char            opc0;	/* First opcode byte */
	char            opc1;	/* Second opcode byte */
	char            modrm;	/* Mode/RM byte */
	struct ex       addr;	/* Operand address */
	struct ex       immed;	/* Immediate data */

	base = ptr->base;
	pos = 0;
	sz = size;		/* capture  size */
	size = 0;
	lsz = 0;
	rsz = 0;
	if (labfld)
		enter(labfld, location, section);
	i = ptr->type;
	addr.mode = 0;
	immed.mode = 0;
	if (i & TLEFT) {
		f = address(&immed);
		lsz = size;
		size = 0;
	}
	if (i & TRIGHT) {
		if (f != 2)
			error('T');
		else {
			f = address(&addr);
			rsz = size;
			size = 0;
		}
	}
	if (i & TVAR) {
		if (f != 2)
			f = 3;
		else {
			f = address(&addr);
			rsz = size;
			size = 0;
		}
	}
	switch (i & 0xfff) {
	case TMODE:
		if(valid(immed.mode, VABS)) {
			switch(immed.val) {
			case 80386:
			case 386:
				mmode = 1;
				break;
			default:
				mmode = 0;
				break;
			}
		} else
			error('r');
		printit();
		return;
	case TZERO1:
		opc0 = base;
		break;
	case TZERO2:
		opc0 = (base >> 8) & 0xff;
		opc1 = base & 0xff;
		flag |= TWOB;
		break;
	case TOPR:
		/* op[s] #v, ax  | 00xxx10w immed */
		/* op[s] #v, ea  | 1000000w yyxxxyyy immed */
		/* op[s] #sv, ea | 1000001w yyxxxyyy immed[8] */
		/* op[s] r, ea   | 00xxx00w yyrrryyy */
		/* op[s] ea, r   | 00xxx01w yyrrryyy */
		sz = checksize(sz, rsz, lsz);
		opc0 = (sz == 1) ? 0 : 1;
		if (sz == 4)
			flag |= LG;
		if (sz == 2)
			flag |= WD;
		if (valid(immed.mode, VIMD)) {
			flag |= IMMED;
			if (valid(addr.mode, VREG) && (addr.mode & 07) == 0) {
				opc0 |= 0x04 | (base << 3);
				break;	/* no! mod/rm byte */
			} else if (valid(addr.mode, VMOD)) {
				opc0 |= 0x80;
				modrm = base;
				if (immed.sym == P_ABS && isbyte(immed.val)) {
					flag |= SIMMED;
					opc0 |= (sz == 1) ? 0 : 2;
				}
			} else
				error('A');
		} else {
			if (valid(immed.mode, VREG) && valid(addr.mode, VMOD)) {
				opc0 |= base << 3;
			} else if (valid(immed.mode, VMOD) && valid(addr.mode, VREG)) {
				swap(&immed, &addr);
				opc0 |= 0x02 | (base << 3);
			} else
				error('A');
			modrm = immed.mode & 07;
		}
checkea:
		if (valid(addr.mode, VI16 | VT16))
			flag |= ADDR16;
		else if (valid(addr.mode, VIND | VINX | VTWO))
			flag |= ADDR32;
		else if (valid(addr.mode, VABS | VREG))
			flag |= (mmode) ? ADDR32 : ADDR16;
		break;
	case TWORD:
		sz = checksize(sz, rsz, lsz);
		if (sz != 2)
			error('S');	/* only word mode */
		sz = 2;
		opc0 = base;
		if (valid(immed.mode, VREG) && valid(addr.mode, VMOD)) {
			modrm = immed.mode & 07;
		} else
			error('A');
		goto checkea;
	case TTWO:
		sz = checksize(sz, rsz, lsz);
		if (sz == 1)
			error('S');	/* No byte mode */
		if (sz == 2)
			flag |= WD;
		else
			flag |= LG;
		opc0 = base;
		swap(&immed, &addr);	/* Backward mem, reg */
		if (valid(immed.mode, VREG) && valid(addr.mode, VMEM)) {
			modrm = immed.mode & 07;
		} else
			error('A');
		goto checkea;
	case TBIT:
		sz = checksize(sz, rsz, lsz);
		if (sz == 1)
			error('S');	/* No byte mode */
		if (sz == 2)
			flag |= WD;
		else
			flag |= LG;
		if (valid(immed.mode, VREG | VIMD) && valid(addr.mode, VMOD)) {
			if (valid(immed.mode, VIMD)) {
				opc0 = 0x0f;
				opc1 = 0xba;
				flag |= TWOB | SIMMED;
				modrm |= (base >> 8) & 07;
			} else {
				opc0 = 0x0f;
				opc1 = base & 0xff;
				flag |= TWOB;
				modrm |= immed.mode & 07;
			}
		} else
			error('A');
		goto checkea;
	case TCALLR:
		/* call  rel	|	e8 rel */
		/* call  reg,ptr|	9a ptr seg */
		/* jmp   rel8	|	eb rel8 */
		/* jmp   rel	|	e9 rel */
		/* jmp   seg,ptr|	ea ptr seg */
		addr.mode = 0;
		immed.mode = 0;
		f = address(&addr);
		if (mmode == 0)
			flag |= ADDR16;
		else
			flag |= ADDR32;
		/* clear the flag bits which have no meaning */
		flag &= ~(LOCK | REP | SEG);
		if (f == 2) {	/* Copy over the segment */
			if (!valid(addr.mode, VABS))
				error('A');
			immed.sym = addr.sym;
			immed.val = addr.val;
			f = address(&addr);
			if (!valid(addr.mode, VABS))
				error('A');
			flag |= FAR;
			opc0 = (base == 0xe8) ? 0x9a : 0xea;
			break;
		}
		opc0 = base;
		/* Just a normal call */
		if ((base == 0xe9) && valid(addr.mode, VABS)) {
			if (pass==1) {
				sz = 1;
				flag &= ~(ADDR16 | ADDR32);
				flag |= SIMMED;
				addchain(location, addr.sym,
					section, 2, (mmode<<4) | 1);
				break;
			}
			if ((addr.sym->flag & N_TYPE) == section) {
				addr.val = addr.val - location;
				addr.sym = P_ABS;
				i = 0xffffff80 & addr.val;
				swap(&addr, &immed);
				sz = (mmode == 0) ? 2 : 4;
				flag &= ~(ADDR16 | ADDR32);
				flag |= IMMED;
				if(i == 0 || i == 0xffffff80) {
					opc0 = 0xeb;
					flag |= SIMMED;
					sz = 1;
					immed.val -= 2;
				} else {
					immed.val -= 5;
				}
			} else {
				addr.mode = (SEGNONE << 12) | MREL;
				addr.val -= 4;
			}
		} else if ((addr.sym->flag & N_TYPE) == section) {
			addr.val = addr.val - location;
			addr.sym = P_ABS;
			swap(&addr, &immed);
			sz = (mmode == 0) ? 2 : 4;
			flag &= ~(ADDR16 | ADDR32);
			flag |= IMMED;
			immed.val -= 5;
		} else if (valid(addr.mode, VABS)) {
			addr.mode = (SEGNONE << 12) | MREL;
			addr.val -= 1;
		} else if (valid(addr.mode, VMOD)) {
			if (mmode == 0)
				flag |= WD;
			else
				flag |= LG;
			opc0 = 0xff;
			modrm = (base == 0xe8) ? 2: 4;
			goto checkea;
		} else 
			error('A');
		break;
	case TCALL:
		/* callf ea	|	ff yy011yyy */
		/* jmpf  ea	|	ff yy101yyy */
		if (mmode == 0)
			flag |= WD;
		else
			flag |= LG;
		swap(&addr,&immed);
		if (!valid(addr.mode, VMOD))
			error('A');
		opc0 = 0xff;
		modrm = base & 07;
		goto checkea;
	case TSTR:
		sz = checksize(sz, rsz, lsz);
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		opc0 = base | ((sz == 1) ? 0 : 1);
		break;
	case TZERO:
		opc0 = base & 0xff;
		if (base & 0x100)
			flag |= LG;
		else
			flag |= WD;
		break;
	case TDEC:
		sz = checksize(sz, 0, lsz);
		swap(&immed, &addr);	/* need it in address */
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		if (!valid(addr.mode, VMOD))
			error('A');
		if (valid(addr.mode, VREG) && sz != 1)
			opc0 = 0x40 | (base << 3) | (addr.mode & 07);
		else {
			opc0 = (sz == 1) ? 0xfe : 0xff;
			modrm = base & 07;
			goto checkea;
		}
		break;

	case TONE:
		sz = checksize(sz, 0, lsz);
		swap(&immed, &addr);	/* need it in address */
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		if (!valid(addr.mode, VMOD))
			error('A');
		opc0 = (sz == 1) ? 0xf6 : 0xf7;
		modrm = base & 07;
		goto checkea;

	case TDIV:
		/* div	ea,ax	  | f6 yy110yyy		 */
		/* idiv	ea,ax	  | f6 yy111yyy		 */
		/* imul	ea,ax	  | f6 yy101yyy		 */
		/* imul	ea,r	  | 0f af yyrrryyy	 */
		/* imul immd,ea,r | 6b yyrrryyy immd	 */
		/* mul  ea,ax	  | f6 yy100yyy		 */
		f = address(&immed);
		lsz = size;
		size = 0;
		if (f != 2) {		/* Assume op, ax */
			addr.mode = (SEGNONE << 12) | MREG | REGAX;
			rsz = (sz)? sz: ((mmode)? 4: 2);
			f = 3;
		} else {
			f = address(&addr);
			rsz = size;
			size = 0;
		}
		sz = checksize(sz, 0, lsz);
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		if (base == 5 && f == 2) {
			struct ex       reg;

			if (sz == 1)
				error('S');
			f = address(&reg);
			if (immed.sym == P_ABS && isbyte(immed.val)) {
				flag |= SIMMED;
				opc0 = 0x6b;
			} else
				opc0 = 0x69;
			if (!valid(reg.mode, VREG))
				error('A');
			flag |= IMMED;
			modrm = reg.mode & 07;
			goto checkea;
		}
		swap(&immed, &addr);	/* op x,ax */
		if (!valid(immed.mode, VREG) && !valid(addr.mode, VMOD))
			error('A');
		if (base == 5 && (immed.mode & 07) != REGAX) {
			if (sz == 1)
				error('S');
			opc0 = 0x0f;
			opc1 = 0xaf;
			flag |= TWOB;
			modrm = immed.mode & 07;
			goto checkea;
		}
		if ((immed.mode & 07) != REGAX)
			error('r');
		opc0 = 0xf6 | ((sz == 1) ? 0 : 1);
		modrm = base;
		goto checkea;
	case TENT:
		if (!valid(immed.mode, VIMD) && !valid(addr.mode, VIMD | VABS))
			error('A');
		genword((long) base, P_ABS, 1, 0);
		if (addr.sym != P_ABS && pass == 2)
			error('R');
		genword(immed.val, immed.sym, 2, 0);
		genword(addr.val, addr.sym, 1, 0);
		printit();
		flag = 0;
		return;
	case TIO:
		/* in immed,ax	| e4 immed */
		/* in dx,ax	| ec */
		/* out ax,immed	| e6 immed */
		/* out ax,dx	| ee */
		if(base == 0xe6) {
			swap(&immed, &addr);	/* op x,ax */
			rsz = lsz;
		}
		sz = checksize(sz, 0, rsz);
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		if (!valid(addr.mode, VREG) && (addr.mode & 07) != REGAX)
			error('r');
		opc0 = base | ((sz == 1) ? 0 : 1);
		flag &= ~(ADDR16 | ADDR32);
 		if (valid(immed.mode, VIMD))
			flag |= SIMMED;
		else if (valid(immed.mode, VREG) && (immed.mode & 07) == REGDX)
			opc0 |= 0x08;
		else
			error('A');
		break;
	case TINT:
		if (!valid(immed.mode, VIMD | VABS))
			error('A');
		if (immed.sym == P_ABS && immed.val == 3)
			opc0 = 0xcc;
		else {
			opc0 = 0xcd;
			flag |= SIMMED;
		}
		break;
	case TSBRA:
		/* jmp   rel8	|	0f 7x rel */
		/* clear the flag bits which have no meaning */
		flag &= ~(LOCK | REP | SEG);
		if (!valid(immed.mode, VABS))
			error('A');
		if ((immed.sym->flag & N_EXT) == 0 &&
		    (immed.sym->flag & N_TYPE) == section) {
			immed.val = immed.val - location;
			immed.sym = P_ABS;
			sz = 4;
			opc0 = base;
			immed.val -= 2;
		} else  error('R');
		break;
	case TBRA:
		/* jcc   rel8	|	7x rel8 */
		/* jcc   rel	|	0f 8x rel */
		/* clear the flag bits which have no meaning */
		flag &= ~(LOCK | REP | SEG);
		if (!valid(immed.mode, VABS))
			error('A');
		if (pass==1) {
			sz = 1;
			flag |= IMMED;
			addchain(location, immed.sym, section, 2, 
				(mmode<<4)|2);
			break;
		}
		if ((immed.sym->flag & N_TYPE) == section) {
			immed.val = immed.val - location;
			immed.sym = P_ABS;
			i = 0xffffff80 & immed.val;
			if(i == 0 || i == 0xffffff80) {
				opc0 = 0x70 | (0xf & base);
				sz = 1;
				flag |= IMMED;
				immed.val -= 2;
				break;
			}
			if(mmode) {
				sz = 4;
				opc0 = 0x0f;
				opc1 = 0x80 | (0xf & base);
				flag |= TWOB | IMMED;
				immed.val -= 6;
			} else {
				opc0 = 0x70 | (0xf & (base ^ 1));
				opc1 = 0x3;
				modrm = 0xea;
				flag |= TWOB | MODRM;
				sz = 2;
				immed.val -= 5;
			}
		} else {
			if(mmode == 0) {
				opc0 = 0x70 | (0xf & (base ^ 1));
				opc1 = 0x3;
				modrm = 0xea;
				flag |= TWOB | MODRM | ADDR16;
				sz = 2;
			} else {
				opc0 = 0x0f;
				opc1 = 0x80 | (0xf & base);
				flag |= TWOB | ADDR32;
			}
			addr.mode = (SEGNONE << 12) | MREL;
			addr.sym = immed.sym;
			addr.val = immed.val - (mmode)? 4: 2;
		}
		break;
	case TLGDT:
		if(valid(immed.mode, VMOD)) {
			swap(&immed, &addr);
			opc0 = 0xf;
			opc1 = 0x1;
			flag |= TWOB;
			modrm = base;
			goto checkea;
		} else
			error('A');
		break;
	case TPRE:
		if (base == 0xf0)
			flag |= LOCK;
		else if (base == 0xf2 || base == 0xf3) {
			flag |= REP;
			repeat = base;
		} else {
			f = address(&immed);
			if (!valid(immed.mode, VREGS))
				error('r');
			segment = immed.mode & 07;
			flag |= SEG;
		}
		printit();
		return;
	case TMOV:
		if (valid(addr.mode, VREGT | VREGD | VREGC) ||
		    valid(immed.mode, VREGT | VREGD | VREGC)) {
			opc0 = 0xf;
			opc1 = 0x20;
			flag |= TWOB | ADDR32;
			if (valid(immed.mode, VREGT | VREGD | VREGC) &&
			    valid(addr.mode, VREG)) {
				opc1 |= 0x2;
			} else if (valid(addr.mode, VREGT | VREGD | VREGC) &&
				   valid(immed.mode, VREG)) {
				swap(&addr, &immed);
			} else
				error('A');
			if (checksize(sz, rsz, lsz) != 4)
				error('s');
			switch (RMODE & immed.mode) {
			case MREGT:
				opc1 |= 4;
				break;
			case MREGD:
				opc1 |= 1;
				break;
			case MREGC:
				break;
			}
			modrm |= immed.mode & 07;
			break;
		}
		sz = checksize(sz, rsz, lsz);
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		if (valid(addr.mode, VREGS) && valid(immed.mode, VMOD)) {
			opc0 = 0x8e;
			swap(&addr, &immed);
			modrm = immed.mode & 7;
			goto checkea;
		}
		if (valid(immed.mode, VREGS) && valid(addr.mode, VMOD)) {
			opc0 = 0x8c;
			modrm = immed.mode & 7;
			goto checkea;
		}
		if (valid(immed.mode, VIMD) && valid(addr.mode, VMOD)) {
			if (valid(addr.mode, VREG)) {
				opc0 = ((sz == 1) ? 0xb0 : 0xb8) + (addr.mode & 07);
				flag |= IMMED;
				break;
			}
			opc0 = (sz == 1) ? 0xc6 : 0xc7;
			flag |= IMMED;
			modrm = 0;
			goto checkea;
		}
		if (valid(immed.mode, VREG) && valid(addr.mode, VMOD)) {
			opc0 = (sz == 1) ? 0x88 : 0x89;
			modrm = immed.mode & 07;
			goto checkea;
		}
		if (valid(addr.mode, VREG) && valid(immed.mode, VMOD)) {
			swap(&addr, &immed);
			opc0 = (sz == 1) ? 0x8a : 0x8b;
			modrm = immed.mode & 07;
			goto checkea;
		}
		error('A');
		break;
	case TEXT:
		/* al,  ir	 */
		/* movsbl	bop,lreg	BY,	0x0fbe	 */
		/* movsbw	bop,wreg	WD,	0x0fbe	 */
		/* movswl	wop,lreg	LG,	0x0fbf	 */
		/* movzbl	bop,lreg	BY,	0x0fb6	 */
		/* movzbw	bop,wreg	WD,	0x0fb6	 */
		/* movzwl	wop,lreg	LG,	0x0fb7	 */
		swap(&immed, &addr);	/* there backwords */
		if (valid(addr.mode, VMOD) && valid(immed.mode, VREG)) {
			opc0 = 0xf;
			opc1 = base;
			if (sz == 4)
				flag |= TWOB | LG;
			else if (sz == 2)
				flag |= TWOB | WD;
			else
				flag |= TWOB;
			switch(sz) {
			case 0:	error('s');	break;
			case 1:	if(rsz != 4 || (lsz != 1 && lsz != 0))
					error('1');
				break;
			case 2:	if(rsz != 2 || (lsz != 1 && lsz != 0))
					error('2');
				break;
			case 4:	if(rsz != 4 || (lsz != 2 && lsz != 0))
					error('3');
				break;
			}
			modrm = immed.mode & 7;
			goto checkea;
		} else
			error('A');
		break;
	case TPOP:
		/* pop   ea	|	0x8f xx000yyy */
		/* pop	 reg	|	0x58+r */
		/* ES, CS, SS, DS, FS, GS */
		/* pop	 sreg	|	07  --  17  1f 0fa1 0fa9 */
		/* push	 ea	|	0xff xx110yyy */
		/* push  reg	|	0x50+r */
		/* push	 immed	|	0x68 i */
		/* push	 simmed |	0x6a si */
		/* push  sreg	|	06  0e  16  1e 0fa0 0fa8 */
		sz = checksize(sz, 0, lsz);	/* figure size out */
		if (sz == 0)
			sz = (mmode) ? 4 : 2;	/* default */
		else if (sz == 1)
			error('s');	/* No BYTES */
		if (valid(immed.mode, VREGS)) {
			opc0 = ((immed.mode & 07) << 3) | (base == 0x8f);
			if (opc0 >= 040) {
				opc1 = opc0 | 0x80;
				opc0 = 0xf;
				flag |= TWOB;
			} else
				opc0 |= 6;
			break;
		}
		if (sz == 4)
			flag |= LG;
		else
			flag |= WD;
		if (valid(immed.mode, VREG)) {
			opc0 = 0x50 | (immed.mode & 07) | ((base == 0x8f) ? 8 : 0);
			break;
		}
		if (base == 0xff && valid(immed.mode, VIMD)) {
			if (immed.sym == P_ABS) {
				opc0 = 0x6a;
				flag |= SIMMED;
			} else
				opc0 = 0x68;
			flag |= IMMED;
			break;
		}
		if (!valid(immed.mode, VMOD))
			error('s');
		else swap(&addr, &immed);
		opc0 = base;
		modrm = (base == 0xff) ? 060 : 0;
		goto checkea;
	case TSHF:
		/* op	#1,ea	|	0xd0 xxyyyxxx */
		/* op	cl,ea	|	0xd2 xxyyyxxx */
		/* op	#z,ea	|	0xc0 xxyyyxxx z */
		/* opD	#z,r,ea	|	0x0f 0xax   xxrrrxxx z */
		/* opD  cl,r,ea	|	0x0f 0xax|1 xxrrrxxx z */
		if(base == 0xfac || base == 0xfa4) {
			struct ex       reg;
			
			if(f != 2) 
				error('T');
			else
				f = address(reg);
			if (lsz != 1 && lsz != 0)
				error('s');
			if (valid(immed.mode, VREG) && (immed.mode & 07) == REGCX) {
				opc0 += 1;
			} else if (valid(immed.mode, VIMD)) {
				flag |= SIMMED;
			} else
				error('A');
			swap(&reg, &addr);	/* Put reg into addr */
			if(valid(reg.mode, VREG))
				modrm = reg.mode & 07;
			else
				error('A');
			sz = checksize(sz, rsz, size);
			if (sz == 4)
				flag |= LG;
			else if (sz == 2)
				flag |= WD;
			else
				error('s');
			goto checkea;
		}
		if (!valid(addr.mode, VMOD))
			error('A');
		if (lsz != 1 && lsz != 0)
			error('s');
		sz = checksize(sz, rsz, 0);
		if (valid(immed.mode, VREG) && (immed.mode & 07) == REGCX) {
			opc0 = 0xd2;
		} else if (valid(immed.mode, VIMD)) {
			if (immed.sym == P_ABS && immed.val == 1) {
				opc0 = 0xd0;
			} else {
				opc0 = 0xc0;
				flag |= SIMMED;
			}
		} else
			error('A');
		if (sz == 4)
			flag |= LG;
		else if (sz == 2)
			flag |= WD;
		modrm = base;
		if (sz == 2 || sz == 4)
			opc0 |= 1;
		goto checkea;
	case TRET:
		sz = 2;
		opc0 = base;
		flag |= IMMED;
		break;
	case TSET:
		sz = checksize(1, 0, lsz);
		if (sz != 1)
			error('s');
		swap(&immed, &addr);
		if (!valid(addr.mode, VMOD))
			error('A');
		opc0 = 0x0f;
		opc1 = base;
		modrm = 0;
		flag &= ~(LG|WD);
		flag |= TWOB;
		goto checkea;
	case TTEST:
		sz = checksize(sz, rsz, lsz);
		if (sz == 4)
			flag |= LG;
		else if (sz == 2)
			flag |= WD;
		if (valid(immed.mode, VIMD)) {
			flag |= IMMED;
			if (valid(addr.mode, VREG) && (addr.mode & 7) == REGAX) {
				opc0 = (sz == 1) ? 0xa8 : 0xa9;
			}
			if (!valid(addr.mode, VMOD))
				error('A');
			opc0 = (sz == 1) ? 0xf6 : 0xf7;
			goto checkea;
		} else if (valid(immed.mode, VREG)) {
			if (!valid(addr.mode, VMOD))
				error('A');
			opc0 = (sz == 1) ? 0x84 : 0x85;
			modrm = immed.mode & 7;
			goto checkea;
		} else
			error('A');
		break;
	case TXCHG:
		sz = checksize(sz, rsz, lsz);
		if (sz == 4)
			flag |= LG;
		else if (sz == 2)
			flag |= WD;
		if (valid(immed.mode, VMOD) && valid(addr.mode, VREG)) {
			swap(&immed, &addr);
		}
		if (!valid(immed.mode, VREG))
			error('A');
		/* check for short form */
		if (valid(immed.mode, VREG) && (immed.mode & 07) == REGAX &&
		    sz != 1 && valid(addr.mode, VREG)) {
			opc0 = 0x90 | (addr.mode & 07);
			break;
		}
		/* can be in either side */
		if (valid(addr.mode, VREG) && (addr.mode & 07) == REGAX &&
		    sz != 1 && valid(immed.mode, VREG)) {
			opc0 = 0x90 | (immed.mode & 07);
			break;
		}
		/* long form */
		opc0 = base | (sz != 1);
		goto checkea;

	case TFLT1:
		/* fpxxx stn,st0  | esc    opc|n */
		/* fpxxx st0,stn  | esc|4  opc|n */
		/* fpxxx stn	  | esc    opc|n */
		opc0 = (base >> 8);
		opc1 = base & 0xff;
		flag |= TWOB;
		if (f == 2) {
			f = address(&addr);
		} else {
			addr.mode = (SEGNONE << 12) | MREGF | 0;
		}
		if(!valid(immed.mode, VREGF) && !valid(addr.mode, VREGF))
			error('r');
		/* st0, stn  where n != 0 */
		if((immed.mode & 07) == 0 && (addr.mode & 07) != 0) {
			swap(&immed, &addr);
			/* The following do not support Reversing */
			switch(base) {
			case 0xd8d0:	/* fcom */
			case 0xded8:	/* fcomp */
			case 0xdd00:	/* fldl */
			case 0xdb05:	/* fldt */
				error('A');
				break;
			case 0xdd03:	/* fstpl */
			case 0xdb07:	/* fstpt */
				break;		/* These need to be reversed */
			default:
				opc0 |= 0x4;
				break;
			}
		} else  {
			/* st0, st0 or stn, st0 */
			if((addr.mode & 07) != 0)
				error('A');
		}
		opc1 |= 07 & immed.mode;
		break;
	case TFLT2:
		/* fpxxx  stn  | esc n */
		if (valid(immed.mode, VREGF)) {
			opc0 = (base >> 8);
			opc1 = (base & 0xff) | (immed.mode & 07);
			flag |= TWOB;
		} else
			error('r');
		break;
	case TFLT3:
		sz = (lsz == 0)? ((mmode)?4:2) : lsz;
		swap(&immed, &addr);	/* need it in address */
		if (sz == 2)
			flag |= WD;
		else if (sz == 4)
			flag |= LG;
		if (!valid(addr.mode, VMOD))
			error('A');
		if(base == 0xdd07 && valid(addr.mode,VREG) &&
			 (addr.mode &7) == REGAX) {
			opc0 = 0xdf;
			opc1 = 0xe0;
			flag &= ~(ADDR16|ADDR32|WD|LG);
			flag |= TWOB;
			break;
		}
		printf("base = %x ", base);
		opc0 = (base >> 8) &0xff;
		printf("opc0 = %x\n", opc0);
		modrm = base & 07;
		goto checkea;
	default:
		error('d');
	}
	/* First generate rep & lock prefixes */
	if (flag & REP)
		genword((long) repeat, P_ABS, 1, 0);
	else if (flag & LOCK)
		genword(0xf0L, P_ABS, 1, 0);
	/* See if we need to escape the addressing mode */
	if (flag & ADDR16 && mmode >= 1)	/* 32 bit mode */
		genword(0x67L, P_ABS, 1, 0);
	if (flag & ADDR32 && mmode == 0)	/* 16 bit mode */
		genword(0x67L, P_ABS, 1, 0);
	/* Now check the data size */
	if (flag & WD && mmode >= 1)	/* 32 bit mode */
		genword(0x66L, P_ABS, 1, 0);
	if (flag & LG && mmode == 0)	/* 16 bit mode */
		genword(0x66L, P_ABS, 1, 0);
	/* Segment override comes next */
	if (flag & SEG)
		genword((long) segover[segment & 07], P_ABS, 1, 0);
	else if (flag & (ADDR16 | ADDR32) &&
		 ((addr.mode >> 12) & 0xf) != SEGNONE)
		genword((long) segover[((addr.mode >> 12) & 07)], P_ABS, 1, 0);
	/* Now the opcodes */
	genword((long) opc0, P_ABS, 1, 0);	/* First opcode byte */
	if (flag & TWOB)
		genword((long) opc1, P_ABS, 1, 0);
	if (flag & FAR) {
		genword(addr.val, addr.sym, (mmode == 0) ? 2 : 4, 0);
		genword(immed.val, immed.sym, 2, 0);
	} else {
		/* Generate a short jump over the next instruction */
		if (flag & SJMP) {
			genword(0xeb, P_ABS, 1, 0);
			genword(sz + 1, P_ABS, 1, 0);
		}
		/* Next Mod/RM and/or Mod/RM,SIB,off */
		if (flag & (ADDR16 | ADDR32))
			genea(modrm, &addr, flag & ADDR32);
		else if (flag & MODRM)
			genword((long) modrm, P_ABS, 1, 0);
		/* Lastly immedate data */
		if (flag & SIMMED)
			genword(immed.val, P_ABS, 1, 0);
		else if (flag & IMMED)
			genword(immed.val, immed.sym, sz, 0);
	}
	/* Finaly done, clean up for next opcode */
	flag = 0;
	printit();
}

/* see if value will fit in a sign extended byte */
int
isbyte(v)
	long            v;
{
	if ((v > -128) && (v < 128))
		return (1);
	return (0);
}

/* Compute and check size */
int
checksize(osize, rsize, lsize)
	int             osize;
	register int    rsize, lsize;
{
	register int i;
/* oprsize [(right<<2)|left] */
	i = (oprsize[(rsize<<3) | lsize]>>(osize<<2))&07;
	if(i == 0) error('s');
	return(i);
}

/* Eval an addressing mode. */

int
address(p)
	register struct ex *p;
{
	register int    x, y, z, seg;

	sp = 0;
	seg = SEGNONE;
	oldptr = -1;
again:
	p->mode = -1;
	p->val = 0;
	p->sym = P_ABS;
	x = lex();
	switch (x) {
	case MOD:		/* Eat a % */
		x = lex();
		if ((stksym[sp]->flag & N_TYPE) != REG)
			error('r');
	case NUMBER:
		if ((stksym[sp]->flag & N_TYPE) != REG)
			goto l;
		if (seg != SEGNONE)
			error('S');
		y = stksym[sp]->value & RVALUE;
		switch (REGFIELD(stksym[sp]->value)) {
		case XREG:	/* Handle st(n) */
			x = lex();
			if(x == RPARN) {
				x = orop();
				if (stksym[0] == P_ABS && stkval[0] >= 0 &&
				    stkval[0] <= 7)  y = (stkval[0] & 7);
				else error('r');
				if(x != LPARN) error(')');
				else x = lex();
			} else	y = 0;
			/* Fall thru */
		case FREG:
			p->mode = (SEGNONE << 12) | MREGF | y;
			size = (mmode) ? 4: 2;
			break;
		case TREG:
			p->mode = (SEGNONE << 12) | MREGT | y;
			size = 4;
			break;
		case DREG:
			p->mode = (SEGNONE << 12) | MREGD | y;
			size = 4;
			break;
		case CREG:
			p->mode = (SEGNONE << 12) | MREGC | y;
			size = 4;
			break;
		case BREG:
			p->mode = (SEGNONE << 12) | MREG | y;
			size = 1;
			break;
		case WREG:
			p->mode = (SEGNONE << 12) | MREG | y;
			size = 2;
			break;
		case LREG:
			p->mode = (SEGNONE << 12) | MREG | y;
			size = 4;
			break;
		case SREG:
			p->mode = (SEGNONE << 12) | MREGS | y;
			x = lex();
			if (x == COLON) {
				seg = y;
				goto again;
			}
			size = (mmode) ? 4: 2;
			unlex(x);
			break;
		}
		x = lex();
		break;
	case DOLLAR:
	case NUM:
		if (seg != SEGNONE)
			error('S');
		x = orop();
		p->val = stkval[0];
		p->sym = stksym[0];
		p->mode = (SEGNONE << 12) | MIMD;
		break;
	case RPARN:
		sp++;
		stkval[0] = 0;
		stksym[0] = P_ABS;
		goto parn;
	default:
l:		unlex(x);
		x = orop();
		switch (x) {
		case COMMA:
		case EOL:
		case -1:
			p->val = stkval[0];
			p->sym = stksym[0];
			p->mode = ((seg != SEGDS) ? (seg << 12) : (SEGNONE << 12)) | MABS;
			break;
		case RPARN:
	parn:
			sp++;
			x = lex();
			if (x == MOD) x = lex();	/* Eat % */
			if (x == NUMBER && ((stksym[sp]->flag & N_TYPE) == REG)) {
				switch (REGFIELD(stksym[sp]->value)) {
				case BREG:
					error('A');
					break;
				case WREG:	/* Sixteen bit addressing */
					x = stkval[sp] & 07;
					if (x != REGBX && x != REGBP && x != REGSI && x != REGDI)
						error('A');
					if (x == REGBP && seg == SEGSS)
						seg = SEGNONE;
					p->mode = MI16 | x;
					break;
				case LREG:
					x = stkval[sp] & 07;
					if ((x == REGBP || x == REGSP) && seg == SEGSS)
						seg = SEGNONE;
					p->mode = MIND | x;
					break;
				}
				x = lex();
				if (x != COMMA)
					goto done;
			} else if (x == COMMA) {
				p->mode = MINX;
			} else
				error('A');
			x = lex();
			if (x == MOD) x = lex();	/* Eat the % */
			if (x == NUMBER && (stksym[sp]->flag & N_TYPE) == REG) {
				switch (RMODE & p->mode) {
				case MI16:
					if (REGFIELD(stksym[sp]->value) != WREG)
						error('A');
					x = stkval[sp] & 07;
					y = p->mode & 07;
					if ((y != REGBX || y != REGBP) &&
					    (x != REGDI || x != REGSI))
						error('A');
					p->mode = MT16 | (x << 3) | y;
					break;
				case MINX:
				case MIND:
					if (REGFIELD(stksym[sp]->value) != LREG)
						error('A');
					x = stkval[sp] & 07;
					p->mode =
					    (((RMODE&p->mode)==MINX)?MINX:MTWO)
						 | (x << 3) | (p->mode & 07);
					x = lex();
					if (x == COMMA) {
						x = lex();
						if (x == NUMBER && stksym[sp] == P_ABS) {
							switch (stkval[sp]) {
							case 1:
								y = 0000;
								break;
							case 2:
								y = 0100;
								break;
							case 4:
								y = 0200;
								break;
							case 8:
								y = 0300;
								break;
							default:
								y = 0;
								error('A');
							}
							p->mode |= y;
						} else
							error('A');
					} else
						unlex(x);
					break;
				default:
					error('A');
				}
				x = lex();
		done:		if (x != LPARN)
					error(')');
				p->val = stkval[0];
				p->sym = stksym[0];
				p->mode |= (seg << 12);
				x = lex();
				break;
		default:
				error('*');
			}
		}
	}
	switch (x) {
	case -1:
		return (0);
	case EOL:
		return (1);
	case COMMA:
		return (2);
	default:
		return (-1);
	}
}


swap(a, b)
	struct ex      *a, *b;

{
	register int    i;
	register long int j;
	register struct symbol *p;

	j = a->val;
	a->val = b->val;
	b->val = j;
	p = a->sym;
	a->sym = b->sym;
	b->sym = p;
	i = a->mode;
	a->mode = b->mode;
	b->mode = i;
}

int
valid(addr, mode)
	unsigned        addr, mode;

{
	switch (addr & RMODE) {
	case MABS:
		return (0 != (mode & VABS));
	case MIND:
		return (0 != (mode & VIND));
	case MTWO:
		return (0 != (mode & VTWO));
	case MINX:
		return (0 != (mode & VINX));
	case MI16:
		return (0 != (mode & VI16));
	case MT16:
		return (0 != (mode & VT16));
	case MIMD:
		return (0 != (mode & VIMD));
	case MREG:
		return (0 != (mode & VREG));
	case MREGT:
		return (0 != (mode & VREGT));
	case MREGD:
		return (0 != (mode & VREGD));
	case MREGC:
		return (0 != (mode & VREGC));
	case MREGS:
		return (0 != (mode & VREGS));
	case MREGF:
		return (0 != (mode & VREGF));
	default:
		return (0);
	}
}

/* Return the size of a variable length instruction based on location
and symbol, extra has size of long instruction */
int
instsize(p)
struct shortop *p;
{
	int i;

	if ((p->sym->flag & N_TYPE) == p->flags) {
		i = 0xffffff80 & (p->sym->value - p->loc);
		if(i == 0 || i == 0xffffff80) 
			return(2);
	}
	switch(p->extra) {
	case 0x11:		/* jmp ; 386 */
		return 5;
	case 0x01:		/* jmp ; real */
		return 3;
	case 0x12:		/* jcc ; 386 */
	case 0x02:		/* jcc ; real */
		return 6;
	default:
		error('I');
		return(p->size);
	}
}


void
genea(reg, p, type)
	char            reg;
	struct ex      *p;
	int             type;
{
	/* ax, cx, dx, bx, sp, bp, si, di */
	static char     remap[8] = {0, 0, 0, 7, 0, 6, 4, 5};
	char            modrm, sib, f;

/* 	fprintf(stderr, "genea(reg= %o, mode = %o, sym = %o, value = %lX, type = %d )\n"
		,reg, p->mode, p->sym, p->val, type);  */
	modrm = (reg & 07) << 3;
	sib = 0;
	switch (p->mode & RMODE) {
	case MREL:
		genword(p->val, p->sym, (type) ? 4 : 2, 1);
		return;
	case MABS:
		genword((long) modrm | ((type) ? 0x05 : 0x06), P_ABS, 1, 0);
		genword(p->val, p->sym, (type) ? 4 : 2, 0);
		return;
	case MIND:
		if ((p->mode & 07) == REGSP) {
			modrm |= 0x04;
			sib = 0x24;
		} else
			modrm |= p->mode & 07;
		goto gen;
	case MTWO:
		modrm |= 0x04;
		sib = p->mode & 0xff;	/* generated correct */
gen:
		if (p->sym == P_ABS) {
			if (p->val != 0L) {
				if (isbyte(p->val))
					modrm |= 0x40;
				else
					modrm |= 0x80;
			}
		} else
			modrm |= 0x80;
		genword((long) modrm, P_ABS, 1, 0);
		if ((modrm & 07) == 0x4)
			genword((long) sib, P_ABS, 1, 0);
		if (modrm & 0x40)
			genword(p->val, P_ABS, 1, 0);
		else if (modrm & 0x80)
			genword(p->val, p->sym, 4, 0);
		return;
	case MINX:
		genword((long) modrm | 0x04, P_ABS, 1, 0);
		genword((long) (p->mode & 0xf8) | 0x05, P_ABS, 1, 0);
		genword(p->val, p->sym, 4, 0);
		return;
	case MI16:
		modrm |= 07&remap[p->mode & 07];
		if ((modrm & 07) == 0)
			error('x');
		goto gen16;
	case MT16:
		if ((p->mode & 07) == REGBP)
			modrm |= 0x2;
		if ((p->mode & 070) == (REGDI << 3))
			modrm |= 0x1;
gen16:
		if (p->sym == P_ABS) {
			if (p->val != 0L) {
				if (isbyte(p->val))
					modrm |= 0x40;
				else
					modrm |= 0x80;
			}
		} else
			modrm |= 0x80;
		genword((long) modrm, P_ABS, 1, 0);
		if (modrm & 0x40)
			genword(p->val, P_ABS, 1, 0);
		else if (modrm & 0x80)
			genword(p->val, p->sym, 2, 0);
		return;
	case MREG:
		genword((long) (modrm | 0xc0 | (p->mode & 07)), P_ABS, 1, 0);
		return;
	case MIMD:
	case MREGT:
	case MREGD:
	case MREGC:
	case MREGS:
		return;
	}
}
