/***************************************************
 *                                                 *
 *    Macro Assembler			           *
 *                                                 *
 *    Copyright 1989    Richard P. Cornwell        *
 *    Developmental Code not for redistubution     *
 *                                                 *
 *    Written by Richard Cornwell                  *
 *               3/20/86                           *
 *                                                 *
 ***************************************************/

#include "macro.h"

#define LG	0x4000
#define WD	0x2000
#define BY	0x1000

struct opcodes optab[] = {
	".align",	EVEN,		1,
	".ascii",	ASCII,		0,
	".asciz",	ASCIZ,		0,
	".bss",		BSS,		0,
	".byte",	BY|BYTE,	0,
	".comm",	COMM,		0,
	".data",	DSECT,		0,
	".db",		BY|BLOCK,	0,
	".dl",		LG|BLOCK,	0,
	".dw",		WD|BLOCK,	0,
	".end",		END,		0,
#ifdef MAXCOND
	".endc",	ENDC,		0,
#endif
	".entry",	ENTRY,		0,
	".equ",		EQU,		0,
	".even",	EVEN,		0,
	".extern",	EXTERN,		0,
	".globl",	GLOBAL,		0,
#ifdef MAXCOND
	".if",		IF,		0,
	".ifd",		IFD,		0,
	".ife",		IFE,		0,
	".iff",		IFF,		0,
	".ifn",		IFN,		0,
	".ift",		IFT,		0,
	".iftf",	IFTF,		0,
	".ifu",		IFU,		0,
#endif
	".lcomm",	LCOMM,		0,
	".long",	LG|LONG,	0,
#ifdef MACASM

	".macro",	MACRO,		0,
	".mend",	MEND,		0,
	".mglb",	MGLB,		0,
	".mgo",		MGO,		0,
	".mlcl",	MLCL,		0,
#endif
	".mode",	TMODE,		0,
#ifdef MACASM
	".mset",	MSET,		0,
	".msetl",	MSETL,		0,
#endif 
	".page",	PAGE,		0,
	".print",	PRINT,		0,
	".set",		SET,		0,
	".space",	SPACE,		0,
	".stabd",	STABD,		0,
	".stabn",	STABN,		0,
	".stabs",	STABS,		0,
	".stitle",	STITLE,		0,
	".text",	CSECT,		0,
	".title",	TITLE,		0,
	".word",	WD|WORD,	0,
	"aaa",		TZERO1,		0x37,
	"aad",		TZERO2,		0xd50a,
	"aam",		TZERO2,		0xd40a,
	"aas",		TZERO1,		0x3f,
	"adc",		TOPR,		0x2,
	"adcb",		BY|TOPR,	0x2,
	"adcl",		LG|TOPR,	0x2,
	"adcw",		WD|TOPR,	0x2,
	"add",		TOPR,		0x0,
	"addb",		BY|TOPR,	0x0,
	"addl",		LG|TOPR,	0x0,
	"addw",		WD|TOPR,	0x0,
	"and",		TOPR,		0x4,
	"andb",		BY|TOPR,	0x4,
	"andl",		LG|TOPR,	0x4,
	"andw",		WD|TOPR,	0x4,
	"arpl",		WD|TWORD,	0x63,
	"ascii",	ASCII,		0,
	"asciz",	ASCIZ,		0,
	"bound",	TTWO,		0x62,
	"boundl",	LG|TTWO,	0x62,
	"boundw",	WD|TTWO,	0x62,
	"bsf",		TTWO,		0x0fbc,
	"bsfl",		LG|TTWO,	0x0fbc,
	"bsfw",		WD|TTWO,	0x0fbc,
	"bsr",		TTWO,		0x0fbd,
	"bsrl",		LG|TTWO,	0x0fbd,
	"bsrw",		WD|TTWO,	0x0fbd,
	"bss",		BSS,		0,
	"bt",		TBIT,		0x04a3,
	"btl",		TBIT,		0x04a3,
	"btw",		TBIT,		0x04a3,
	"btc",		TBIT,		0x07bb,
	"btcl",		LG|TBIT,	0x07bb,
	"btcw",		WD|TBIT,	0x07bb,
	"btr",		TBIT,		0x06b3,
	"btrl",		LG|TBIT,	0x06b3,
	"btrw",		WD|TBIT,	0x05ab,
	"bts",		TBIT,		0x05ab,
	"btsl",		LG|TBIT,	0x05ab,
	"btsw",		WD|TBIT,	0x05ab,
	"byte",		BY|BYTE,	0,
	"call",		TCALLR,		0xe8,
	"callf",	TCALL,		0x03,
	"calli",	TCALL,		0x02,
	"cbw",		WD|TZERO,	0x0098,
	"cdq",		LG|TZERO,	0x0199,
	"clc",		TZERO1,		0xf8,
	"cld",		TZERO1,		0xfc,
	"cli",		TZERO1,		0xfa,
	"clts",		TZERO2,		0x0f06,
	"cmc",		TZERO1,		0xf5,
	"cmp",		TOPR,		0x7,
	"cmpb",		BY|TOPR,	0x7,
	"cmpl",		LG|TOPR,	0x7,
	"cmps",		TSTR,		0xa6,
	"cmpsb",	BY|TSTR,	0xa6,
	"cmpsl",	LG|TSTR,	0xa6,	
	"cmpsw",	WD|TSTR,	0xa6,
	"cmpw",		WD|TOPR,	0x7,
	"comm",		COMM,		0,
	"cwd",		WD|TZERO,	0x0099,
	"cwde",		LG|TZERO,	0x0198,
	"daa",		TZERO1,		0x27,
	"das",		TZERO1,		0x2f,
	"data",		DSECT,		0,
	"dec",		TDEC,		0x01,
	"decb",		BY|TDEC,	0x01,
	"decl",		LG|TDEC,	0x01,	
	"decw",		WD|TDEC,	0x01,
	"div",		TDIV,		0x6,
	"divb",		BY|TDIV,	0x6,
	"divl",		LG|TDIV,	0x6,
	"divw",		WD|TDIV,	0x6,
	"ds",		BLOCK,		0,
	"end",		END,		0,
#ifdef MAXCOND
	"endc",		ENDC,		0,
#endif
	"enter",	TENT,		0xc8,
	"entry",	ENTRY,		0,
	"equ",		EQU,		0,
	"even",		EVEN,		0,
	"extern",	EXTERN,		0,
	"f2xm1",	TZERO2,		0xd9f0,
	"fabs",		TZERO2,		0xd9e1,
	"fadd",		TFLT1,		0xd8c0,
	"faddl",	TFLT3,		0xdc00,
	"faddp",	TFLT1,		0xdac0,
	"fadds",	TFLT3,		0xd800,
	"fbld",		TFLT3,		0xdf04,
	"fbstp",	TFLT3,		0xdf06,
	"fchs",		TZERO2,		0xd9e0,
	"fclex",	TZERO2,		0xdde2,
	"fcom",		TFLT1,		0xd8d0,
	"fcoml",	TFLT3,		0xdc02,
	"fcomp",	TFLT1,		0xded8,
	"fcompl",	TFLT3,		0xdc03,
	"fcompp",	TZERO2,		0xded9,
	"fcomps",	TFLT3,		0xd803,
	"fcoms",	TFLT3,		0xd802,
	"fcos",		TZERO2,		0xd9ff,
	"fdecstp",	TZERO2,		0xd9f6,
	"fdiv",		TFLT1,		0xd8f0,
	"fdivl",	TFLT3,		0xdc06,
	"fdivp",	TFLT1,		0xdaf0,
	"fdivr",	TFLT1,		0xd8f8,
	"fdivrl",	TFLT3,		0xdc07,
	"fdivrp",	TFLT1,		0xdaf8,
	"fdivrs",	TFLT3,		0xd807,
	"fdivs",	TFLT3,		0xd806,
	"ffree",	TFLT2,		0xddc0,
	"fiaddl",	TFLT3,		0xda00,
	"fiadds",	TFLT3,		0xde00,
	"ficoml",	TFLT3,		0xda02,
	"ficompl",	TFLT3,		0xda03,
	"ficomps",	TFLT3,		0xde03,
	"ficoms",	TFLT3,		0xde02,
	"fidivl",	TFLT3,		0xda06,
	"fidivrl",	TFLT3,		0xda07,
	"fidivrs",	TFLT3,		0xde07,
	"fidivs",	TFLT3,		0xde06,
	"fildl",	TFLT3,		0xdb00,
	"fildq",	TFLT3,		0xdf05,
	"filds",	TFLT3,		0xdf00,
	"fimull",	TFLT3,		0xda01,
	"fimuls",	TFLT3,		0xde01,
	"fincstp",	TZERO2,		0xd9f7,
	"finit",	TZERO2,		0xdde2,
	"fistl",	TFLT3,		0xdb02,
	"fistpl",	TFLT3,		0xdb03,
	"fistpq",	TFLT3,		0xdf07,
	"fistps",	TFLT3,		0xdf03,
	"fists",	TFLT3,		0xdf02,
	"fisubl",	TFLT3,		0xda04,
	"fisubrl",	TFLT3,		0xda05,
	"fisubrs",	TFLT3,		0xde05,
	"fisubs",	TFLT3,		0xde04,
	"fld",		TFLT2,		0xd9c0,
	"fld1",		TZERO2,		0xd9e8,
	"fldcw",	TFLT3,		0xd905,
	"fldenv",	TFLT3,		0xd904,
	"fldl",		TFLT3,		0xdd00,
	"fldl2e",	TZERO2,		0xd9ea,
	"fldl2t",	TZERO2,		0xd9e9,
	"fldlg2",	TZERO2,		0xd9ec,
	"fldln2",	TZERO2,		0xd9ed,
	"fldpi",	TZERO2,		0xd9eb,
	"flds",		TFLT3,		0xd900,
	"fldt",		TFLT1,		0xdb05,
	"fldz",		TZERO2,		0xd9ee,
	"fmul",		TFLT1,		0xd8c8,
	"fmull",	TFLT3,		0xdc01,
	"fmulp",	TFLT1,		0xdac8,
	"fmuls",	TFLT3,		0xd801,
	"fnop",		TZERO2,		0xd9d0,
	"fnstsw",	TFLT3,		0xdd07,
	"fpatan",	TZERO2,		0xd9f3,
	"fprem",	TZERO2,		0xd9f8,
	"fprem1",	TZERO2,		0xd9f5,
	"fptan",	TZERO2,		0xd9f2,
	"frndint",	TZERO2,		0xd9fc,
	"frstor",	TFLT3,		0xdd04,
	"fsave",	TFLT3,		0xdd06,
	"fscale",	TZERO2,		0xd9fd,
	"fsin",		TZERO2,		0xd9fe,
	"fsincos",	TZERO2,		0xd9fd,
	"fsqrt",	TZERO2,		0xd9fa,
	"fst",		TFLT2,		0xddd0,
	"fstcw",	TFLT3,		0xdd07,
	"fstenv",	TFLT3,		0xd906,
	"fstl",		TFLT3,		0xdd02,
	"fstp",		TFLT2,		0xddd8,
	"fstpl",	TFLT3,		0xdd03,
	"fstps",	TFLT3,		0xd903,
	"fstpt",	TFLT3,		0xdb07,
	"fsts",		TFLT3,		0xd902,
	"fstsw",	TFLT3,		0xdd07,
	"fsub",		TFLT1,		0xd8e0,
	"fsubl",	TFLT3,		0xdc04,
	"fsubp",	TFLT1,		0xdae0,
	"fsubr",	TFLT1,		0xd8e8,
	"fsubrl",	TFLT3,		0xdc05,
	"fsubrp",	TFLT1,		0xdae8,
	"fsubrs",	TFLT3,		0xd805,
	"fsubs",	TFLT3,		0xd804,
	"ftst",		TZERO2,		0xd9e4,
	"fucom",	TFLT2,		0xdde0,
	"fucomp",	TFLT2,		0xdde8,
	"fucompp",	TZERO2,		0xdae9,
	"fxam",		TZERO2,		0xd9e5,
	"fxch",		TFLT2,		0xd9c8,
	"fxtract",	TZERO2,		0xd9f4,
	"fyl2x",	TZERO2,		0xd9f1,
	"fyl2xp1",	TZERO2,		0xd9f9,
	"globl",	GLOBAL,		0,
	"hlt",		TZERO,		0xf4,
	"idiv",		TDIV,		0x7,
	"idivb",	BY|TDIV,	0x7,
	"idivl",	LG|TDIV,	0x7,
	"idivw",	WD|TDIV,	0x7,
#ifdef MAXCOND
	"if",		IF,		0,
	"ifd",		IFD,		0,
	"ife",		IFE,		0,
	"iff",		IFF,		0,
	"ifn",		IFN,		0,
	"ift",		IFT,		0,
	"iftf",		IFTF,		0,
	"ifu",		IFU,		0,
#endif
	"imul",		TDIV,		0x5,
	"imulb",	BY|TDIV,	0x5,
	"imull",	LG|TDIV,	0x5,
	"imulw",	WD|TDIV,	0x5,
	"in",		TIO,		0xe4,
	"inb",		BY|TIO,		0xe4,
	"inc",		TDEC,		0x00,
	"incb",		BY|TDEC,	0x00,
	"incl",		LG|TDEC,	0x00,
	"incw",		WD|TDEC,	0x00,
	"inl",		LG|TIO,		0xe4,
	"insb",		BY|TSTR,	0x6c,
	"insl",		LG|TSTR,	0x6c,
	"insw",		WD|TSTR,	0x6c,
	"int",		TINT,		0xcc,
	"into",		TZERO1,		0xce,
	"inw",		WD|TIO,		0xe4,
	"iret",		TZERO1,		0xcf,
	"iretd",	TZERO1,		0xcf,
	"ja",		TBRA,		0x77,
	"jae",		TBRA,		0x73,
	"jb",		TBRA,		0x72,
	"jbe",		TBRA,		0x76,
	"jc",		TBRA,		0x72,
	"jcxz",		TBRA,		0xe3,
	"je",		TBRA,		0x74,
	"jecxz",	TBRA,		0xe3,
	"jg",		TBRA,		0x7f,
	"jge",		TBRA,		0x7d,
	"jl",		TBRA,		0x7c,
	"jle",		TBRA,		0x7e,
	"jmp",		TCALLR,		0xe9,
	"jmpf",		TCALL,		0x05,
	"jmpi",		TCALL,		0x04,
	"jna",		TBRA,		0x76,
	"jnae",		TBRA,		0x72,
	"jnb",		TBRA,		0x73,
	"jnbe",		TBRA,		0x77,
	"jnc",		TBRA,		0x73,
	"jne",		TBRA,		0x75,
	"jng",		TBRA,		0x7e,
	"jnge",		TBRA,		0x7c,
	"jnl",		TBRA,		0x7d,
	"jnle",		TBRA,		0x7f,
	"jno",		TBRA,		0x71,
	"jnp",		TBRA,		0x7b,
	"jns",		TBRA,		0x79,
	"jnz",		TBRA,		0x75,
	"jo",		TBRA,		0x70,
	"jp",		TBRA,		0x7a,
	"jpe",		TBRA,		0x7a,
	"jpo",		TBRA,		0x7b,
	"js",		TBRA,		0x78,
	"jz",		TBRA,		0x74,
	"lahf",		TZERO1,		0x9f,
	"lar",		TWORD,		0x0f02,
	"larl",		LG|TWORD,	0x0f02,
	"larw",		WD|TWORD,	0x0f02,
	"lcomm",	LCOMM,		0,
	"lds",		TWORD,		0xc5,
	"ldsl",		LG|TWORD,	0xc5,
	"ldsw",		WD|TWORD,	0xc5,
	"lea",		TTWO,		0x8d,
	"leal",		LG|TTWO,	0x8d,
	"leave",	TZERO1,		0xc9,
	"leaw",		WD|TTWO,	0x8d,
	"les",		TWORD,		0xc4,
	"lesl",		LG|TWORD,	0xc4,
	"lesw",		WD|TWORD,	0xc4,
	"lfs",		TWORD,		0x0fb4,
	"lfsl",		LG|TWORD,	0x0fb4,
	"lfsw",		WD|TWORD,	0x0fb4,
	"lgdt",		TLGDT,		0x0102,
	"lgdtl",	LG|TLGDT,	0x0102,
	"lgdtw",	WD|TLGDT,	0x0102,
	"lgs",		TWORD,		0x0fb5,
	"lgsl",		LG|TWORD,	0x0fb5,
	"lgsw",		WD|TWORD,	0x0fb5,
	"lidt",		TLGDT,		0x0103,
	"lidtl",	LG|TLGDT,	0x0103,
	"lidtw",	WD|TLGDT,	0x0103,
	"lldt",		TLGDT,		0x0002,
	"lmsw",		TLGDT,		0x0106,
	"lock",		TPRE,		0xf0,
	"lodsb",	BY|TSTR,	0xac,
	"lodsl",	LG|TSTR,	0xac,
	"lodsw",	WD|TSTR,	0xac,
	"long",		LONG,		0,
	"loop",		TBRA,		0xe2,
	"loope",	TBRA,		0xe1,
	"loopz",	TBRA,		0xe1,
	"loopne",	TBRA,		0xe0,
	"loopnz",	TBRA,		0xe0,
	"lsl",		TWORD,		0x0f03,
	"lsll",		LG|TWORD,	0x0f03,
	"lslw",		WD|TWORD,	0x0f03,
	"lss",		TWORD,		0x0fb2,
	"lssl",		LG|TWORD,	0x0fb2,
	"lssw",		WD|TWORD,	0x0fb2,
	"ltr",		TLGDT,		0x0003,
	"ltrw",		WD|TLGDT,	0x0003,
#ifdef MACASM
	"macro",	MACRO,		0,
	"mend",		MEND,		0,
	"mglb",		MGLB,		0,
	"mgo",		MGO,		0,
	"mlcl",		MLCL,		0,
#endif
	"mov",		TMOV,		0x0,
	"movb",		BY|TMOV,	0x0,
	"movl",		LG|TMOV,	0x0,
	"movsb",	BY|TSTR,	0xa4,
	"movsbl",	BY|TEXT,	0x0fbe,
	"movsbw",	WD|TEXT,	0x0fbe,
	"movsl",	LG|TSTR,	0xa4,
	"movsw",	WD|TSTR,	0xa4,
	"movswl",	LG|TEXT,	0x0fbf,	
	"movw",		WD|TMOV,	0x0,
	"movzbl",	BY|TEXT,	0x0fb6,
	"movzbw",	WD|TEXT,	0x0fb6,
	"movzwl",	LG|TEXT,	0x0fb7,
#ifdef MACASM
	"mset",		MSET,		0,
	"msetl",	MSETL,		0,
#endif 
	"mul",		TDIV,		0x4,
	"mulb",		BY|TDIV,	0x4,
	"mull",		LG|TDIV,	0x4,
	"mulw",		WD|TDIV,	0x4,
	"neg",		TONE,		0x3,
	"negb",		BY|TONE,	0x3,
	"negl",		LG|TONE,	0x3,
	"negw",		WD|TONE,	0x3,
	"nop",		TZERO1,		0x90,
	"not",		TONE,		0x2,
	"notb",		BY|TONE,	0x2,
	"notl",		LG|TONE,	0x2,
	"notw",		WD|TONE,	0x2,
	"or",		TOPR,		0x1,
	"orb",		BY|TOPR,	0x1,
	"orl",		LG|TOPR,	0x1,
	"orw",		WD|TOPR,	0x1,
	"out",		TIO,		0xe6,
	"outb",		BY|TIO,		0xe6,
	"outl",		LG|TIO,		0xe6,
	"outsb",	BY|TSTR,	0x6e,
	"outsl",	LG|TSTR,	0x6e,
	"outsw",	WD|TSTR,	0x6e,
	"outw",		WD|TIO,		0xe6,
	"page",		PAGE,		0,
	"pop",		TPOP,		0x8f,
	"popa",		TZERO1,		0x61,
	"popad",	TZERO1,		0x61,
	"popf",		TZERO1,		0x9d,
	"popfd",	TZERO1,		0x9d,
	"popl",		LG|TPOP,	0x8f,
	"popw",		WD|TPOP,	0x8f,
	"print",	PRINT,		0,
	"push",		TPOP,		0xff,
	"pusha",	TZERO1,		0x60,
	"pushan",	TZERO1,		0x60,
	"pushf",	TZERO1,		0x9c,
	"pushfd",	TZERO1,		0x9c,
	"pushl",	LG|TPOP,	0xff,
	"pushw",	WD|TPOP,	0xff,
	"rcl",		TSHF,		0x2,
	"rclb",		BY|TSHF,	0x2,
	"rcll",		LG|TSHF,	0x2,
	"rclw",		WD|TSHF,	0x2,
	"rcr",		TSHF,		0x3,
	"rcrb",		BY|TSHF,	0x3,
	"rcrl",		LG|TSHF,	0x3,
	"rcrw",		WD|TSHF,	0x3,
	"rep",		TPRE,		0xf3,
	"repe",		TPRE,		0xf3,
	"repne",	TPRE,		0xf2,
	"ret",		TZERO1,		0xc3,
	"retd",		TRET,		0xc2,
	"retdf",	TRET,		0xca,
	"retf",		TZERO1,		0xcb,
	"rol",		TSHF,		0x0,
	"rolb",		BY|TSHF,	0x0,
	"roll",		LG|TSHF,	0x0,
	"rolw",		TSHF,		0x0,
	"ror",		TSHF,		0x1,
	"rorb",		BY|TSHF,	0x1,
	"rorl",		LG|TSHF,	0x1,
	"rorw",		WD|TSHF,	0x1,
	"sahf",		TZERO1,		0x9e,
	"sal",		TSHF,		0x4,
	"salb",		BY|TSHF,	0x4,
	"sall",		LG|TSHF,	0x4,
	"salw",		WD|TSHF,	0x4,
	"sar",		TSHF,		0x7,
	"sarb",		BY|TSHF,	0x7,
	"sarl",		LG|TSHF,	0x7,
	"sarw",		WD|TSHF,	0x7,
	"sbb",		TOPR,		0x3,
	"sbbb",		BY|TOPR,	0x3,
	"sbbl",		LG|TOPR,	0x3,
	"sbbw",		WD|TOPR,	0x3,
	"scasb",	BY|TSTR,	0xae,
	"scasl",	LG|TSTR,	0xae,
	"scasw",	WD|TSTR,	0xae,
	"set",		SET,		0,
	"seta",		TSET,		0x0f97,
	"setae",	TSET,		0x0f93,
	"setb",		TSET,		0x0f92,
	"setbe",	TSET,		0x0f96,
	"setc",		TSET,		0x0f92,
	"sete",		TSET,		0x0f94,
	"setg",		TSET,		0x0f9f,
	"setge",	TSET,		0x0f9d,
	"setl",		TSET,		0x0f9c,
	"setle",	TSET,		0x0f9e,
	"setna",	TSET,		0x0f96,
	"setnae",	TSET,		0x0f92,
	"setnb",	TSET,		0x0f93,
	"setnbe",	TSET,		0x0f97,
	"setnc",	TSET,		0x0f93,
	"setne",	TSET,		0x0f95,
	"setng",	TSET,		0x0f9e,
	"setnge",	TSET,		0x0f9c,
	"setnl",	TSET,		0x0f9d,
	"setnle",	TSET,		0x0f9f,
	"setno",	TSET,		0x0f91,
	"setnp",	TSET,		0x0f9b,
	"setns",	TSET,		0x0f99,
	"setnz",	TSET,		0x0f95,
	"seto",		TSET,		0x0f90,
	"setp",		TSET,		0x0f9a,
	"setpe",	TSET,		0x0f9a,
	"setpo",	TSET,		0x0f9b,
	"sets",		TSET,		0x0f98,
	"setz",		TSET,		0x0f94,
	"sgdt",		TLGDT,		0x0100,
	"shl",		TSHF,		0x4,
	"shlb",		BY|TSHF,	0x4,
	"shld",		TSHF,		0x0fa4,
	"shldl",	LG|TSHF,	0x0fa4,
	"shldw",	WD|TSHF,	0x0fa4,
	"shll",		LG|TSHF,	0x4,
	"shlw",		WD|TSHF,	0x4,
	"shr",		TSHF,		0x5,
	"shrb",		BY|TSHF,	0x5,
	"shrd",		TSHF,		0x0fac,
	"shrdl",	LG|TSHF,	0x0fac,
	"shrdw",	WD|TSHF,	0x0fac,
	"shrl",		LG|TSHF,	0x5,
	"shrw",		WD|TSHF,	0x5,
	"sidt",		TLGDT,		0x0101,
	"sldt",		TLGDT,		0x0000,
	"smsw",		TLGDT,		0x0104,
	"space",	SPACE,		0,
	"stabd",	STABD,		0,
	"stabn",	STABN,		0,
	"stabs",	STABS,		0,
	"stc",		TZERO1,		0xf9,
	"std",		TZERO1,		0xfd,
	"sti",		TZERO1,		0xfb,
	"stitle",	STITLE,		0,
	"stosb",	BY|TSTR,	0xaa,
	"stosl",	LG|TSTR,	0xaa,
	"stosw",	WD|TSTR,	0xaa,
	"str",		TLGDT,		0x0001,
	"sub",		TOPR,		0x5,
	"subb",		BY|TOPR,	0x5,
	"subl",		LG|TOPR,	0x5,
	"subw",		WD|TOPR,	0x5,
	"test",		TTEST,		0xa8,
	"testb",	BY|TTEST,	0xa8,
	"testl",	LG|TTEST,	0xa8,
	"testw",	WD|TTEST,	0xa8,
	"text",		CSECT,		0,
	"title",	TITLE,		0,
	"verr",		TLGDT,		0x0004,
	"verw",		TLGDT,		0x0005,
	"wait",		TZERO1,		0x9b,
	"word",		WORD,		0,
	"xchg",		TXCHG,		0x90,
	"xchgb",	BY|TXCHG,	0x90,
	"xchgl",	LG|TXCHG,	0x90,
	"xchgw",	WD|TXCHG,	0x90,
	"xlat",		TZERO1,		0xd7,
	"xor",		TOPR,		0x6,
	"xorb",		BY|TOPR,	0x6,
	"xorl",		LG|TOPR,	0x6,
	"xorw",		WD|TOPR,	0x6,
	};

int opsize = sizeof(optab)/sizeof(struct opcodes);

#define WR (WREG<<8)
#define BR (BREG<<8)
#define LR (LREG<<8)
#define SR (SREG<<8)
#define CR (CREG<<8)
#define DR (DREG<<8)
#define TR (TREG<<8)
#define FR (FREG<<8)

struct opcodes reglist[] = {
	"ax",		WR|0,		REG,
	"bx",		WR|3,		REG,
	"cx",		WR|1,		REG,
	"dx",		WR|2,		REG,
	"si",		WR|6,		REG,
	"di",		WR|7,		REG,
	"bp",		WR|5,		REG,
	"sp",		WR|4,		REG,
	"eax",		LR|0,		REG,
	"ebx",		LR|3,		REG,
	"ecx",		LR|1,		REG,
	"edx",		LR|2,		REG,
	"esi",		LR|6,		REG,
	"edi",		LR|7,		REG,
	"ebp",		LR|5,		REG,
	"esp",		LR|4,		REG,
	"al",		BR|0,		REG,
	"ah",		BR|4,		REG,
	"bl",		BR|3,		REG,
	"bh",		BR|7,		REG,
	"cl",		BR|1,		REG,
	"ch",		BR|5,		REG,
	"dl",		BR|2,		REG,
	"dh",		BR|6,		REG,
	"cs",		SR|1,		REG,
	"ss",		SR|2,		REG,
	"ds",		SR|3,		REG,
	"es",		SR|0,		REG,
	"fs",		SR|4,		REG,
	"gs",		SR|5,		REG,
	"cr0",		CR|0,		REG,	
	"cr1",		CR|1,		REG,
	"cr2",		CR|2,		REG,
	"cr3",		CR|3,		REG,
	"dr0",		DR|0,		REG,
	"dr1",		DR|1,		REG,
	"dr2",		DR|2,		REG,
	"dr3",		DR|3,		REG,
	"dr4",		DR|5,		REG,
	"dr5",		DR|5,		REG,
	"dr6",		DR|6,		REG,
	"dr7",		DR|7,		REG,
	"tr6",		TR|6,		REG,
	"tr7",		TR|7,		REG,
	"st",		(XREG<<8)|0,	REG,
	"st0",		FR|0,		REG,
	"st1",		FR|1,		REG,
	"st2",		FR|2,		REG,
	"st3",		FR|3,		REG,
	"st4",		FR|4,		REG,
	"st5",		FR|5,		REG,
	"st6",		FR|6,		REG,
	"st7",		FR|7,		REG,
	NULL,		0,		0 };
