/***************************************************
 *                                                 *
 *    Macro Assembler			           *
 *                                                 *
 *    Written by Richard Cornwell                  *
 *               3/20/86                           *
 *                                                 *
 ***************************************************/

#include "macro.h"

#define REGFIELD(x)	((x>>8)&0377)

machine(ptr)	struct opcodes *ptr; 
{
	register int i, sz, flag;
	unsigned int base, mode;
	struct ex right, left;

	base = ptr->base;
	pos = 0;
	sz = size;
	if(labfld) enter(labfld,location,section);
	i = ptr->type;
	left.mode = 0;
	right.mode = 0;
	switch(i) {
	case TBIT:
		break;
	case TBRA:
		flag = address(&left);
		if(flag != 2) base |= 0x0800;
		else {
			if(left.mode != MCON) error('C');
			base |= (left.val&0xf)<<8;
			flag = address(&left);
			}
		if(left.mode!= MDIR)	error('A');
		left.val = (left.val - (location+2)) >>1 ;
		if((left.sym->flag&N_TYPE) != section && pass == 2) 
				error('R');
		if(pass == 2 && (left.val>-128) && (left.val<127)) error('B');
		base |= left.val&0xff;
		left.mode = MREG;
		break;
	case TCMP:
		break;
	case TDIE:
		break;
	case TDJN:
		break;
	case TFLG:
		break;
	case TINC:
		flag = address(&left);
		if(flag != 2) {
			right.mode = MIMD;
			right.val = 1;
			right.sym = P_ABS;
			}
		else		flag = address(&right);
		if(!valid(right.mode,VIMD) || !valid(left.mode,VDST))error('A');
		if(base&0x0100) {
			if(valid(left.mode,VREG) && 
				REGFIELD(left.sym->value) != WREG)error('A');
			sz = 2;
			}
		else {
			if(valid(left.mode,VREG) && 
				REGFIELD(left.sym->value) != BREG)error('A');
			sz = 1;
			}
		if(pass==2 && (right.sym->flag&N_TYPE) != N_ABS) error('A');
		if(pass==2 && right.val <= 0 || right.val > 16) error('R');
		base |= ((left.mode&0x30)<<10) | ((left.mode&0xf)<<4);
		base |= (right.val-1)&0xf;
		right.mode = MREG;
		break;
	case TJMP:
		flag = address(&left);
		if(flag == 2) {
			if(base == 0x1f00)	error('T');
			else {
				if(left.mode != MCON) error('C');
				base |= left.val&0xf;
				flag = address(&left);
				}
			}
		else {
			if(base != 0x1f00)	base |= 0x8;
			}
		if(!valid(left.mode,(~VREG)&VDST))	error('A');
		base |= ((left.mode&0x30)<<10)| ((left.mode&0xf)<<4);
		break;
	case TLDA:
		break;
	case TLDC:
		break;
	case TLDM:
		break;
	case TLDR:
		break;
	case TLDX:
		flag = address(&left);
		if(flag != 2)	error('T');
		else		flag = address(&right);
		if(base&0x1000) sz = 3;
		else { if(base&0x0100) sz = 2;
			else  sz = 1;
			}
		if(valid(left.mode,VREG)) {
			if(valid(right.mode,VBAS)) {
				if(sz==3) base = 0x7500;
				else 	  base |= 0x7000;
				base |= ((right.mode&0xf)<<4);
				right.val = (right.val&0xf0)<<4;
				right.mode = MDIR;
				right.sym = P_ABS;
				}
			else if(!valid(right.mode,VSRC)) error('A');
			else {
				if(valid(right.mode,VIMD) &&
				   valid(left.mode,VREG) &&
				   (REGFIELD(left.sym->value)==BREG)) {
					base = 0xc000 | ((left.mode&0xf)<<8);
					base |= (right.val&0xff);
					right.mode = VREG;
					break;
					}
				}
			}
		else if(valid(left.mode,VDST)) {
			if(valid(right.mode,VREG)) {
				if(sz==3) base |= 0x0900;
				else	  base |= 0x0e00;
				swap(&left, &right);
				}
			else if(valid(right.mode,VIMD)) {
				if(sz==3)	error('A');
				else if(sz==2)	base = 0x0d05;
				else		base = 0x0c05;
				swap(&left,&right);
				}
			}
		else if(valid(left.mode,VBAS)) {
			if(sz==3) base = 0x7700;
			else 	  base |= 0x7200;
			swap(&left, &right);
			base |= ((right.mode&0xf)<<4);
			right.val = (right.val&0xf0)<<4;
			right.mode = MDIR;
			right.sym = P_ABS;
			}
		base |= ((right.mode&0x30)<<10) | ((right.mode&0xf)<<4);
		base |= left.mode&0xf;
		break;
	case TMOV:
		break;
	case TONE:
		flag = address(&left);
		if(flag == 2)	error('T');
		if(!valid(left.mode,VDST))	error('A');
		base |= ((left.mode&0x30)<<10)| ((left.mode&0xf)<<4);
		break;
	case TPOP:
		flag = address(&left);
		if(flag != 2)	error('T');
		else		flag = address(&right);
		if(base&0x0400) swap(&right,&left);
		if(!valid(left.mode,VIND) || !valid(right.mode,VSRC))
			error('A');
		sz = (base&0x0200)? 2 : 3;
		base |= ((right.mode&0x30)<<10)| (right.mode&0xf);
		base |= ((left.mode&0xf)<<4);
		break;
	case TREG:
		flag = address(&left);
		if(flag != 2)	error('T');
		else		flag = address(&right);
		if(!valid(left.mode,VREG) || !valid(right.mode,VREG))
			error('A');
		if(base&0x0100) {
			if((REGFIELD(left.sym->value)!=WREG) ||
			   (REGFIELD(right.sym->value)!=WREG)) error('A');
			}
		else	{
			if((REGFIELD(left.sym->value)!=BREG) ||
			   (REGFIELD(right.sym->value)!=BREG)) error('A');
			}
		base |= ((left.mode&0xf)<<4) | (right.mode&0xf);
		break;
	case TREL:
		break;
	case TRET:
		break;
	case TROL:
		break;
	case TSC:
		flag = address(&left);
		if(flag == 2)	error('T');
		if(left.mode != MIMD)	error('A');
		if(left.sym->flag != N_ABS && pass == 2) error('A');
		left.mode = 0;
		base |= (left.val&0xff);
		break;
	case TSGN:
		break;
	case TSHF:
		break;
	case TTWO:
		flag = address(&left);
		if(flag != 2)	error('T');
		else		flag = address(&right);
		if(!valid(left.mode,VREG))	error('A');
		if(base&0x1000) {
			if(REGFIELD(left.sym->value) != LREG) error('A');
			sz = 3;
		} else {
			if(base&0x0100) {
				if(REGFIELD(left.sym->value) != WREG)error('A');
				sz = 2;
				}
			else {
				if(REGFIELD(left.sym->value) != BREG)error('A');
				sz = 1;
				}
			}
		if(!valid(right.mode,VSRC))	error('A');
		base |= ((right.mode&0x30)<<10) | ((right.mode&0xf)<<4);
		base |= left.mode&0xf;
		break;
	case TZER:
		break;
	}
	genword(base,P_ABS,2);
	genea(&left,sz);
	genea(&right,sz);
	printit();
}


/* Eval an addressing mode. */

int address(p)
register struct ex *p;
{ register int x, y, z;

	sp = 0;
	oldptr = -1;
	p->val = 0;
	p->sym = P_ABS;
	x = lex();
	switch(x) {
	case NUM:
		x = orop();
		p->val = stkval[0];
		p->sym = stksym[0];
		p->mode = MIMD;
		break;
	case AT:
	case RPARN:
		y = lex();
		if(y == NUMBER && ((stksym[sp]->flag&N_TYPE) == REG)) {
			p->mode = MIND| (stkval[0]&017);
			if(x == RPARN && lex() != LPARN) error(')');
			x = lex();
			break;
			}
		unlex(y);
		goto l;
	case NUMBER:
		if((stksym[sp]->flag&N_TYPE) == REG) {
			p->val = stkval[sp];
			p->sym = stksym[sp];
			switch((stksym[sp]->value>>8)&0377) {
			case CREG:
				p->mode = MCON;
				break;
			case SREG:
			case IREG:
				p->mode = MSPC;
				break;
			case WREG:
			case BREG:
			case LREG:
			case QREG:
				p->mode = MREG|(p->val&017);
				break;
			}
			x = lex();
			if(x == RPARN) {
				y = lex();
				if(y == NUMBER) {
					p->mode = MBAS| (p->mode&017);
					p->val = ((stkval[0]&017)<<4) |
						(p->mode&017);
					if(lex() != LPARN) error(')');
					x = lex();
					}
				}
			break;
		}
		/* fall thru */
	default:
	l:	unlex(x);
		x = orop();
		switch(x) {
		case COMMA:
		case EOL:
		case -1:
			p->val = stkval[0];
			p->sym = stksym[0];
			p->mode = MDIR;
			break;
		case RPARN:
			sp++;
			y = lex();
			if(y != NUMBER &&
			   ((stksym[sp]->value>>8)&0377) != WREG ) {
				error('A');
				}
			if(lex() != LPARN) error('*');
			p->val = stkval[0];
			p->sym = stksym[0];
			p->mode = MINX | (stkval[sp]&0xf);
			x = lex();
			break;
		default: error('*');
		}
	}
	switch(x) {
		case -1: return(0);
		case EOL: return(1);
		case COMMA: return(2);
		default: return(-1);
		}
	}

swap(a,b)
struct ex *a, *b;
{
	register int i;
	register struct symbol *p;
	
	i=a->val;
	a->val = b->val;
	b->val = i;
	p=a->sym;
	a->sym = b->sym;
	b->sym = p;
	i=a->mode;
	a->mode = b->mode;
	b->mode = i;
	}

int valid(addr,mode) unsigned addr, mode;
{
	switch(addr&RMODE) {
		case MIND:	return((0!=(mode&VIND)) && ((mode&017) != 0));
		case MINX:	return((0!=(mode&VINX)) && ((mode&017) != 0));
		case MREG:	return(0!=(mode&VREG));
		case MCON:	return(0!=(mode&VCON));
		case MIMD:	return(0!=(mode&VIMD));
		case MDIR:	return(0!=(mode&VDIR));
		case MBAS:	return(0!=(mode&VBAS));
		case MBRG:	return(0!=(mode&VBRG));
		case MSPC:	return(0!=(mode&VSPC));
		}
}

int genea(p,sz) struct ex *p; unsigned sz;
{
	switch(p->mode&RMODE) {
	case MIND:
	case MREG:
	case MCON:
	case MSPC:
	case MBRG:
		return;	/* nothing to do */
	case MINX:
	case MDIR:
		genword(p->val,p->sym,4);
		return;
	case MIMD:
		switch(sz) {
		case 3:		genword(p->val,p->sym,4);
				break;
		case 2:		genword(p->val,p->sym,2);
				break;
		case 1:		genword(((p->val&0xff)<<8)| (p->val&0xff),
					P_ABS, 2);
				break;
		}
		return;
	case MBAS:
		genword((p->val&017)<<8,P_ABS,2);
		return;
	}
}

